﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Primitives;

namespace Platform.Logging
{
    public abstract class PlatformBatchLoggerProvider<TLoggerOptions> : ILoggerProvider
        where TLoggerOptions : PlatformBatchLoggerOptions, new()
    {
        #region Construction

        private Task _OutputTask;
        private IDisposable _OptionsMonitor;
        private BlockingCollection<PlatformBatchLoggerMessage> _MessageQueue;
        private CancellationTokenSource _CancellationTokenSource;
        private ConcurrentDictionary<string, PlatformBatchLogger<TLoggerOptions>> _Loggers;

        private static Func<string, LogLevel, bool> DefaultFilter = (n, l) => l >= LogLevel.Information;

        private static Dictionary<LogLevel, Func<string, LogLevel, bool>> LevelFilters = new Dictionary<LogLevel, Func<string, LogLevel, bool>> {
            { LogLevel.None, (category,level) => false },
            { LogLevel.Trace, (category,level) => level >= LogLevel.Trace },
            { LogLevel.Debug, (category,level) => level >= LogLevel.Debug },
            { LogLevel.Information, (category,level) => level >= LogLevel.Information },
            { LogLevel.Warning, (category,level) => level >= LogLevel.Warning },
            { LogLevel.Error, (category,level) => level >= LogLevel.Error },
            { LogLevel.Critical, (category,level) => level >= LogLevel.Critical }
        };

        protected PlatformBatchLoggerProvider( IConfigurationSection configuration )
        {
            // get configuration section (Generic Logging)
            Configuration = configuration;

            // create loggers collection
            _Loggers = new ConcurrentDictionary<string, PlatformBatchLogger<TLoggerOptions>>();

            // signal configuration changed
            OnConfigurationChanged();

            // process change
            _OptionsMonitor = ChangeToken.OnChange(configuration.GetReloadToken, OnConfigurationChanged);

            // create message queue
            _MessageQueue = new BlockingCollection<PlatformBatchLoggerMessage>( ProviderOptions.BackgroundQueueSize ?? LoggingOptions.BackgroundQueueSize ?? 4096 );

            // create cancellation token source
            _CancellationTokenSource = new CancellationTokenSource();

            // create output task
            _OutputTask = Task.Run( () => OutputMessagesAsync(_CancellationTokenSource.Token) );
        }
        
        ~PlatformBatchLoggerProvider()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose( bool disposing )
        {
            if( disposing )
            {
                // cancel logger
                _CancellationTokenSource.Cancel();

                // quit option monitor
                _OptionsMonitor.Dispose();
                _OptionsMonitor = null;

                // wait for task to finish
                _OutputTask.Wait();

                // dispose all related items
                _CancellationTokenSource.Dispose();
                _MessageQueue.Dispose();
                _OutputTask.Dispose();
            }
        }

        #endregion

        #region Properties

        protected abstract string SectionName { get; }
        public IConfigurationSection Configuration { get; private set; }
        public PlatformBatchLoggerOptions LoggingOptions { get; private set; }
        public TLoggerOptions ProviderOptions { get; private set; }

        #endregion

        #region Create Loggers

        protected virtual void OnConfigurationChanged()
        {
            LoggingOptions = new PlatformBatchLoggerOptions();
            Configuration.Bind(LoggingOptions);

            ProviderOptions = new TLoggerOptions();
            Configuration.GetSection(SectionName).Bind(ProviderOptions);

            foreach (var logger in _Loggers.Values)
                logger.Filter = GetFilter( logger.Category );
        }

        internal abstract PlatformBatchLogger<TLoggerOptions> CreateLoggerImplementation( string categoryName );

        public ILogger CreateLogger(string categoryName)
        {
            return _Loggers.GetOrAdd( categoryName, CreateLoggerImplementation( categoryName ) );
        }

        #endregion

        #region Output Messages

        /// <summary>
        /// Add logging message
        /// </summary>
        /// <param name="message"></param>
        internal void AddMessage(PlatformBatchLoggerMessage message)
        {
            try
            {
                if (!_MessageQueue.IsAddingCompleted)
                    _MessageQueue.TryAdd( message );
            }
            catch (Exception ex)
            {
                SimpleFileLogger.WriteLine(Configuration, "{0} Logger failed to add message: {1}", SectionName, ex.Message);
            }
        }

        private DateTime LastCleanup;

        /// <summary>
        /// Output message
        /// </summary>
        /// <param name="cancellationToken"></param>
        /// <returns></returns>
        private async Task OutputMessagesAsync( CancellationToken cancellationToken )
        {
            while( !cancellationToken.IsCancellationRequested )
            {
                // get first message into batch
                var messages = new List<PlatformBatchLoggerMessage>();
                if (_MessageQueue.TryTake(out var message, (int)TimeSpan.FromHours(12).TotalMilliseconds, cancellationToken))
                {
                    // add initial message
                    messages.Add(message);

                    // get other available messages in batch
                    for (var limit = ProviderOptions?.BatchSize ?? LoggingOptions?.BatchSize ?? int.MaxValue;
                        limit > 0 && _MessageQueue.TryTake(out message, 0, cancellationToken);
                        limit--) messages.Add(message);

                    // check if canceled
                    if (messages.Count > 0 && !cancellationToken.IsCancellationRequested)
                        await WriteMessagesAsync(messages, cancellationToken);
                }

                if (cancellationToken.IsCancellationRequested)
                    break;

                int retention = ProviderOptions.RetentionInDays ?? LoggingOptions.RetentionInDays ?? 90;
                if (retention > 0 && LastCleanup.Date != DateTime.Now.Date)
                {
                    LastCleanup = DateTime.Now;
                    await CleanupLoggingAsync(retention);
                }

                var delay = ProviderOptions?.FlushPeriod ?? LoggingOptions?.FlushPeriod ?? TimeSpan.FromSeconds(4);
                await Task.Delay(delay, cancellationToken);
            }
        }

        protected virtual Task CleanupLoggingAsync( int retention )
        {
            return Task.CompletedTask;
        }

        protected abstract Task WriteMessagesAsync(IEnumerable<PlatformBatchLoggerMessage> messages, CancellationToken token);

        #endregion

        #region Tools

        protected string GetLogLevelString(LogLevel logLevel)
        {
            switch (logLevel)
            {
                case LogLevel.Trace:
                    return "trce";
                case LogLevel.Debug:
                    return "dbug";
                case LogLevel.Information:
                    return "info";
                case LogLevel.Warning:
                    return "warn";
                case LogLevel.Error:
                    return "fail";
                case LogLevel.Critical:
                    return "crit";
                default:
                    throw new ArgumentOutOfRangeException(nameof(logLevel));
            }
        }

        private IEnumerable<string> GetKeyPrefixes(string category)
        {
            while (!string.IsNullOrEmpty(category))
            {
                yield return category;

                var lastIndexOfDot = category.LastIndexOf('.');

                if (lastIndexOfDot == -1)
                {
                    yield return "Default";
                    break;
                }

                category = category.Substring(0, lastIndexOfDot);
            }
        }

        internal Func<string, LogLevel, bool> GetFilter(string category)
        {
            var section1 = Configuration.GetSection(SectionName + ":LogLevel");
            var section2 = Configuration.GetSection("LogLevel");

            foreach (var prefix in GetKeyPrefixes(category))
            {
                var value = section1.GetValue<string>(prefix);
                if (string.IsNullOrEmpty(value)) value = section2.GetValue<string>(prefix);

                LogLevel level;
                if (!Enum.TryParse<LogLevel>(value, true, out level)) continue;

                return LevelFilters[level];
            }

            return DefaultFilter;
        }

        #endregion
    }
}
